from contoml import tokens
from contoml.tokens import toml2py
from contoml.elements import common
from contoml.elements.common import Element
from contoml.elements.errors import InvalidElementError

_opening_bracket_types = (tokens.TYPE_OP_SQUARE_LEFT_BRACKET, tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET)
_closing_bracket_types = (tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET)
_name_types = (
    tokens.TYPE_BARE_STRING,
    tokens.TYPE_LITERAL_STRING,
    tokens.TYPE_STRING,
)


class TableHeaderElement(Element):
    """
    An element containing opening and closing single and double square brackets, strings and dots and ending with
    a newline.
    
    Raises InvalidElementError.
    """
    
    def __init__(self, _tokens):
        Element.__init__(self, common.TYPE_MARKUP)
        TableHeaderElement._validate_tokens(_tokens)
        self._tokens = _tokens
        self._names = tuple(toml2py.deserialize(token) for token in self._tokens if token.type in _name_types)

    @property
    def is_array_of_tables(self):
        opening_bracket = next(token for i, token in enumerate(self._tokens) if token.type in _opening_bracket_types)
        return opening_bracket.type == tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET

    @property
    def names(self):
        """
        Returns a sequence of string names making up this table header name.
        """
        return self._names

    def has_name_prefix(self, names):
        """
        Returns True if the header names is prefixed by the given sequence of names.
        """
        for i, name in enumerate(names):
            if self.names[i] != name:
                return False
        return True

    def serialized(self):
        return ''.join(token.source_substring for token in self._tokens)

    def is_named(self, names):
        """
        Returns True if the given name sequence matches the full name of this header.
        """
        return tuple(names) == self.names

    @staticmethod
    def _validate_tokens(_tokens):

        opening_bracket_i = next((i for i, token in enumerate(_tokens)
                                  if token.type in _opening_bracket_types), float('-inf'))

        if opening_bracket_i < 0:
            raise InvalidElementError('Expected an opening bracket')

        _tokens = _tokens[opening_bracket_i+1:]
        first_name_i = next((i for i, token in enumerate(_tokens) if token.type in _name_types), float('-inf'))
        if first_name_i < 0:
            raise InvalidElementError('Expected a table header name')

        _tokens = _tokens[first_name_i+1:]

        while True:

            next_dot_i = next((i for i, token in enumerate(_tokens) if token.type == tokens.TYPE_OPT_DOT),
                              float('-inf'))
            if next_dot_i < 0:
                break

            _tokens = _tokens[next_dot_i+1:]

            next_name_i = next((i for i, token in enumerate(_tokens) if token.type in _name_types), float('-inf'))
            if next_name_i < 0:
                raise InvalidElementError('Expected a name after the dot')

            _tokens = _tokens[next_name_i+1:]

        closing_bracket_i = next((i for i, token in enumerate(_tokens) if token.type in _closing_bracket_types),
                                 float('-inf'))

        if closing_bracket_i < 0:
            raise InvalidElementError('Expected a closing bracket')

        if _tokens[-1].type != tokens.TYPE_NEWLINE:
            raise InvalidElementError('Must end with a newline')



from contoml.util import is_sequence_like, is_dict_like

def test_is_sequence_like():
    assert is_sequence_like([1, 3, 4])
    assert not is_sequence_like(42)
    

def test_is_dict_like():
    assert is_dict_like({'name': False})
    assert not is_dict_like(42)
    assert not is_dict_like([4, 8, 15])



from contoml.file.peekableit import PeekableIterator


def test_peekable_iterator():

    peekable = PeekableIterator(i for i in (1, 2, 3, 4))

    assert peekable.peek() == 1
    assert peekable.peek() == 1
    assert peekable.peek() == 1

    assert [next(peekable), next(peekable), next(peekable), next(peekable)] == [1, 2, 3, 4]





def parse_token_stream(token_stream):
    """
    Parses the given token_stream into a sequence of top-level TOML elements.
    """
    from .parser import toml_file_elements
    from .elementsanitizer import sanitize

    elements, _ = toml_file_elements(token_stream)

    return sanitize(elements)



import re
from datetime import datetime
import pytz
import strict_rfc3339
from contoml import tokens
from contoml.tokens import TYPE_BOOLEAN, TYPE_INTEGER, TYPE_FLOAT, TYPE_DATE, \
    TYPE_MULTILINE_STRING, TYPE_BARE_STRING, TYPE_MULTILINE_LITERAL_STRING, TYPE_LITERAL_STRING, \
    TYPE_STRING
import codecs
import six


def deserialize(token):
    """
    Deserializes the value of a single tokens.Token instance based on its type.
    """
    
    if token.type == TYPE_BOOLEAN:
        return _to_boolean(token)
    elif token.type == TYPE_INTEGER:
        return _to_int(token)
    elif token.type == TYPE_FLOAT:
        return _to_float(token)
    elif token.type == TYPE_DATE:
        return _to_date(token)
    elif token.type in (TYPE_STRING, TYPE_MULTILINE_STRING, TYPE_BARE_STRING,
                        TYPE_LITERAL_STRING, TYPE_MULTILINE_LITERAL_STRING):
        return _to_string(token)
    else:
        raise Exception('This should never happen!')


def _unescape_str(unescaped):
    if six.PY2:
        return unescaped.decode('string-escape').decode('unicode-escape')
    else:
        return codecs.decode(unescaped, 'unicode-escape')


def _to_string(token):
    if token.type == tokens.TYPE_BARE_STRING:
        return token.source_substring

    elif token.type == tokens.TYPE_STRING:
        escaped = token.source_substring[1:-1]
        return _unescape_str(escaped)

    elif token.type == tokens.TYPE_MULTILINE_STRING:
        escaped = token.source_substring[3:-3]
        if escaped[0] == '\n':
            escaped = escaped[1:]

        # Remove all occurrences of a slash-newline-one-or-more-whitespace patterns
        escaped = re.sub('\\\\\n\\s*', repl='', string=escaped, flags=re.DOTALL)
        return _unescape_str(escaped)

    elif token.type == tokens.TYPE_LITERAL_STRING:
        return token.source_substring[1:-1]

    elif token.type == tokens.TYPE_MULTILINE_LITERAL_STRING:
        text = token.source_substring[3:-3]
        if text[0] == '\n':
            text = text[1:]
        return text

    raise RuntimeError('Control should never reach here.')


def _to_int(token):
    return int(token.source_substring.replace('_', ''))

def _to_float(token):
    assert token.type == tokens.TYPE_FLOAT
    string = token.source_substring.replace('_', '')
    return float(string)

def _to_boolean(token):
    return token.source_substring == 'true'

def _to_date(token):
    return datetime.fromtimestamp(strict_rfc3339.rfc3339_to_timestamp(token.source_substring), tz=pytz.utc)



import operator
from functools import reduce
from contoml.file import raw


class CascadeDict:
    """
    A dict-like object made up of one or more other dict-like objects where querying for an item cascade-gets
    it from all the internal dicts in order of their listing, and setting an item sets it on the first dict listed.
    """

    def __init__(self, *internal_dicts):
        assert internal_dicts, 'internal_dicts cannot be empty'
        self._internal_dicts = tuple(internal_dicts)

    def cascaded_with(self, one_more_dict):
        """
        Returns another instance with one more dict cascaded at the end.
        """
        return CascadeDict(self._internal_dicts, one_more_dict,)

    def __getitem__(self, item):
        for d in self._internal_dicts:
            try:
                return d[item]
            except KeyError:
                pass
        raise KeyError

    def __setitem__(self, key, value):
        self._internal_dicts[0][key] = value

    def keys(self):
        return set(reduce(operator.or_, (set(d.keys()) for d in self._internal_dicts)))

    def items(self):
        all_items = reduce(operator.add, (list(d.items()) for d in reversed(self._internal_dicts)))
        unique_items = {k: v for k, v in all_items}.items()
        return tuple(unique_items)

    def __contains__(self, item):
        for d in self._internal_dicts:
            if item in d:
                return True
        return False

    @property
    def neutralized(self):
        return {k: raw.to_raw(v) for k, v in self.items()}



from contoml import tokens, lexer
from contoml.elements import traversal
from contoml.elements.atomic import AtomicElement
from contoml.elements.metadata import NewlineElement, PunctuationElement, WhitespaceElement, CommentElement
from contoml.elements.table import TableElement
from contoml.elements.tableheader import TableHeaderElement

atomic_token_types = (
    tokens.TYPE_INTEGER,
    tokens.TYPE_FLOAT,
    tokens.TYPE_BARE_STRING,
    tokens.TYPE_STRING,
    tokens.TYPE_LITERAL_STRING,
    tokens.TYPE_MULTILINE_STRING,
    tokens.TYPE_MULTILINE_LITERAL_STRING,
)

punctuation_token_types = (
    tokens.TYPE_OPT_DOT,
    tokens.TYPE_OP_CURLY_LEFT_BRACKET,
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET,
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET,
    tokens.TYPE_OP_CURLY_RIGHT_BRACKET,
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
    tokens.TYPE_OP_ASSIGNMENT,
)

def primitive_token_to_primitive_element(token):
    if token.type == tokens.TYPE_NEWLINE:
        return NewlineElement((token,))
    elif token.type in atomic_token_types:
        return AtomicElement((token,))
    elif token.type == tokens.TYPE_NEWLINE:
        return NewlineElement((token,))
    elif token.type in punctuation_token_types:
        return PunctuationElement((token,))
    elif token.type == tokens.TYPE_WHITESPACE:
        return WhitespaceElement((token,))
    elif token.type == tokens.TYPE_COMMENT:
        return CommentElement((token,))
    else:
        raise RuntimeError("{} has no mapped primitive element".format(token))


def primitive_tokens_to_primitive_elements(tokens):
    return list(map(primitive_token_to_primitive_element, tokens))


def dummy_file_elements():
        tokens_ = tuple(lexer.tokenize("""
name = fawzy
another_name=another_fawzy

[details]
id= 42
section =fourth

[[person]]
personname= lefawzy
dest=north

[[person]]
dest=south
personname=lafawzy

[details.extended]
number = 313
type =complex"""))

        elements = \
            [TableElement(primitive_tokens_to_primitive_elements(tokens_[:12]))] + \
            [TableHeaderElement(tokens_[12:16])] + \
            [TableElement(primitive_tokens_to_primitive_elements(tokens_[16:25]))] + \
            [TableHeaderElement(tokens_[25:31])] + \
            [TableElement(primitive_tokens_to_primitive_elements(tokens_[31:39]))] + \
            [TableHeaderElement(tokens_[39:45])] + \
            [TableElement(primitive_tokens_to_primitive_elements(tokens_[45:53]))] + \
            [TableHeaderElement(tokens_[53:60])] + \
            [TableElement(primitive_tokens_to_primitive_elements(tokens_[60:]))]

        return elements


class DummyFile(traversal.TraversalMixin):

    @property
    def elements(self):
        return dummy_file_elements()



import pytest
from contoml import lexer
from contoml.elements.array import ArrayElement
from contoml.elements.atomic import AtomicElement
from contoml.elements.metadata import PunctuationElement, WhitespaceElement, NewlineElement


def test_array_element():
    tokens = tuple(lexer.tokenize('[4, 8, [42, \n 23], 15]'))
    assert len(tokens) == 19
    sub_elements = (
        PunctuationElement(tokens[:1]),

        AtomicElement(tokens[1:2]),
        PunctuationElement(tokens[2:3]),
        WhitespaceElement(tokens[3:4]),

        AtomicElement(tokens[4:5]),
        PunctuationElement(tokens[5:6]),
        WhitespaceElement(tokens[6:7]),

        ArrayElement((
            PunctuationElement(tokens[7:8]),

            AtomicElement(tokens[8:9]),
            PunctuationElement(tokens[9:10]),
            WhitespaceElement(tokens[10:11]),
            NewlineElement(tokens[11:12]),
            WhitespaceElement(tokens[12:13]),

            AtomicElement(tokens[13:14]),
            PunctuationElement(tokens[14:15]),
        )),

        PunctuationElement(tokens[15:16]),
        WhitespaceElement(tokens[16:17]),
        AtomicElement(tokens[17:18]),
        PunctuationElement(tokens[18:19])
    )

    array_element = ArrayElement(sub_elements)

    # Test length
    assert len(array_element) == 4

    # Test getting a value
    assert array_element[0] == 4
    assert array_element[1] == 8
    assert array_element[2][0] == 42
    assert array_element[2][1] == 23
    assert array_element[-1] == 15

    # Test assignment with a negative index
    array_element[-1] = 12

    # Test persistence of formatting
    assert '[4, 8, [42, \n 23], 12]' == array_element.serialized()

    # Test raises IndexError on invalid index
    with pytest.raises(IndexError) as _:
        print(array_element[5])

    # Test appending a new value
    array_element[2].append(77)
    assert '[4, 8, [42, \n 23, 77], 12]' == array_element.serialized()

    # Test deleting a value
    del array_element[2][1]
    assert '[4, 8, [42, 77], 12]' == array_element.serialized()

    # Test primitive_value
    assert [4, 8, [42, 77], 12] == array_element.primitive_value



from contoml.errors import NoArrayFoundError, InvalidValueError
from contoml.file import structurer, entries, raw
from contoml.file.array import ArrayOfTables
from contoml.file.freshtable import FreshTable
import contoml.elements.factory as element_factory


class TOMLFile:
    """
    A TOMLFile object that tries its best to prserve formatting and order of mappings of the input source.

    Raises InvalidTOMLFileError on invalid input elements.
    """

    def __init__(self, _elements):
        self._elements = []
        self._navigable = {}
        self.append_elements(_elements)

    def __getitem__(self, item):
        try:
            value = self._navigable[item]
            if isinstance(value, (list, tuple)):
                return ArrayOfTables(parent=self, name=item, iterable=value)
            else:
                return value
        except KeyError:
            return FreshTable(parent=self, name=item, is_array=False)

    def __setitem__(self, key, value):

        if key and isinstance(value, (tuple, list)) and all(isinstance(v, dict) for v in value):
            for table in value:
                self.array(key).append(table)

        elif isinstance(value, dict):
            if key in self._navigable:
                index = self._elements.index(self._navigable[key])
                self._elements = self._elements[:index] + [element_factory.create_table(value)] + self._elements[index+1:]
            else:
                if key:
                    self._elements.append(element_factory.create_table_header_element(key))
                self._elements.append(element_factory.create_table(value))
                self._elements.append(element_factory.create_newline_element())

        else:
            raise InvalidValueError('Assigned value must be a dict or a sequence of dicts')

        self._recreate_navigable()

    def _recreate_navigable(self):
        if self._elements:
            self._navigable = structurer.structure(entries.extract(self._elements))

    def array(self, name):
        """
        Returns the array of tables with the given name.
        """
        if name in self._navigable:
            if isinstance(self._navigable[name], (list, tuple)):
                return self[name]
            else:
                raise NoArrayFoundError
        else:
            return ArrayOfTables(parent=self, name=name)

    def append_elements(self, elements):
        """
        Appends more elements to the contained internal elements.
        """
        self._elements = self._elements + list(elements)
        self._recreate_navigable()

    def dumps(self):
        """
        Returns the TOML file serialized back to str.
        """
        return ''.join(element.serialized() for element in self._elements)

    def dump(self, file_path):
        with open(file_path, mode='w') as fp:
            fp.write(self.dumps())

    def keys(self):
        return self._navigable.keys()

    def values(self):
        return self._navigable.values()

    def items(self):
        return self._navigable.items()

    @property
    def primitive(self):
        """
        Returns a primitive object representation for this container (which is a dict).

        WARNING: The returned container does not contain any markup or formatting metadata.
        """
        return raw.to_raw(self._navigable)

    def append_fresh_table(self, fresh_table):
        """
        Gets called by FreshTable instances when they get written to.
        """
        elements = []
        if fresh_table.name:
            if fresh_table.is_array:
                elements += [element_factory.create_array_of_tables_header_element(fresh_table.name)]
            else:
                elements += [element_factory.create_table_header_element(fresh_table.name)]

        elements += [fresh_table, element_factory.create_newline_element()]
        self.append_elements(elements)




"""
TOML lexical tokens.
"""

class TokenType:
    """
    A TokenType is a concrete type of a source token along with a defined priority and a higher-order kind.

    The priority will be used in determining the tokenization behaviour of the lexer in the following manner:
    whenever more than one token is recognizable as the next possible token and they are all of equal source
    length, this priority is going to be used to break the tie by favoring the token type of the lowest priority
    value. A TokenType instance is naturally ordered by its priority.
    """

    def __init__(self, name, priority, is_metadata):
        self._priority = priority
        self._name = name
        self._is_metadata = is_metadata

    @property
    def is_metadata(self):
        return self._is_metadata

    @property
    def priority(self):
        return self._priority

    def __repr__(self):
        return "{}-{}".format(self.priority, self._name)

    def __lt__(self, other):
        return isinstance(other, TokenType) and self._priority < other.priority

# Possible types of tokens
TYPE_BOOLEAN = TokenType('boolean', 0, is_metadata=False)
TYPE_INTEGER = TokenType('integer', 0, is_metadata=False)
TYPE_OP_COMMA = TokenType('comma', 0, is_metadata=True)
TYPE_OP_SQUARE_LEFT_BRACKET = TokenType('square_left_bracket', 0, is_metadata=True)
TYPE_OP_SQUARE_RIGHT_BRACKET = TokenType('square_right_bracket', 0, is_metadata=True)
TYPE_OP_CURLY_LEFT_BRACKET = TokenType('curly_left_bracket', 0, is_metadata=True)
TYPE_OP_CURLY_RIGHT_BRACKET = TokenType('curly_right_bracket', 0, is_metadata=True)
TYPE_OP_ASSIGNMENT = TokenType('assignment', 0, is_metadata=True)
TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET = TokenType('double_square_left_bracket', 0, is_metadata=True)
TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET = TokenType('double_square_right_bracket', 0, is_metadata=True)
TYPE_FLOAT = TokenType('float', 1, is_metadata=False)
TYPE_DATE = TokenType('date', 40, is_metadata=False)
TYPE_OPT_DOT = TokenType('dot', 40, is_metadata=True)
TYPE_BARE_STRING = TokenType('bare_string', 50, is_metadata=False)
TYPE_STRING = TokenType('string', 90, is_metadata=False)
TYPE_MULTILINE_STRING = TokenType('multiline_string', 90, is_metadata=False)
TYPE_LITERAL_STRING = TokenType('literal_string', 90, is_metadata=False)
TYPE_MULTILINE_LITERAL_STRING = TokenType('multiline_literal_string', 90, is_metadata=False)
TYPE_NEWLINE = TokenType('newline', 91, is_metadata=True)
TYPE_WHITESPACE = TokenType('whitespace', 93, is_metadata=True)
TYPE_COMMENT = TokenType('comment', 95, is_metadata=True)


def is_operator(token):
    """
    Returns True if the given token is an operator token.
    """
    return token.type in (
        TYPE_OP_COMMA,
        TYPE_OP_SQUARE_LEFT_BRACKET,
        TYPE_OP_SQUARE_RIGHT_BRACKET,
        TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
        TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
        TYPE_OP_CURLY_LEFT_BRACKET,
        TYPE_OP_CURLY_RIGHT_BRACKET,
        TYPE_OP_ASSIGNMENT,
        TYPE_OPT_DOT,
    )

class Token:
    """
    A token/lexeme in a TOML source file.

    A Token instance is naturally ordered by its type.
    """

    def __init__(self, _type, source_substring, col=None, row=None):
        self._source_substring = source_substring
        self._type = _type
        self._col = col
        self._row = row

    def __eq__(self, other):
        if not isinstance(other, Token):
            return False
        return self.source_substring == other.source_substring and self.type == other.type

    @property
    def col(self):
        """
        Column number (1-indexed).
        """
        return self._col

    @property
    def row(self):
        """
        Row number (1-indexed).
        """
        return self._row

    @property
    def type(self):
        """
        One of of the TOKEN_TYPE_* constants.
        """
        return self._type

    @property
    def source_substring(self):
        """
        The substring of the initial source file containing this token.
        """
        return self._source_substring

    def __lt__(self, other):
        return isinstance(other, Token) and self.type < other.type

    def __repr__(self):
        return "{}: {}".format(self.type, self.source_substring)



from contoml import lexer
from contoml.elements.atomic import AtomicElement
from contoml.elements.inlinetable import InlineTableElement
from contoml.elements.metadata import PunctuationElement, WhitespaceElement


def test_inline_table():
    tokens = tuple(lexer.tokenize('{ name= first, id=42}'))

    elements = (
        PunctuationElement(tokens[:1]),
        WhitespaceElement(tokens[1:2]),
        AtomicElement(tokens[2:3]),
        PunctuationElement(tokens[3:4]),
        WhitespaceElement(tokens[4:5]),
        AtomicElement(tokens[5:6]),
        PunctuationElement(tokens[6:7]),
        WhitespaceElement(tokens[7:8]),
        AtomicElement(tokens[8:9]),
        PunctuationElement(tokens[9:10]),
        AtomicElement(tokens[10:11]),
        PunctuationElement(tokens[11:12])
    )

    table = InlineTableElement(elements)

    assert table['name'] == 'first'
    assert table['id'] == 42

    table['name'] = 'fawzy'
    table['nickname'] = 'nickfawzy'

    assert set(table.items()) == {('name', 'fawzy'), ('id', 42), ('nickname', 'nickfawzy')}

    assert table.serialized() == '{ name= fawzy, id=42, nickname = nickfawzy}'

    del table['name']

    assert table.serialized() == '{ id=42, nickname = nickfawzy}'

    del table['nickname']

    assert table.serialized() == '{ id=42}'

    del table['id']

    assert table.serialized() == '{ }'

    table['item1'] = 11
    table['item2'] = 22

    assert table.serialized() == '{ item1 = 11, item2 = 22}'



from contoml import elements
from contoml.elements.table import TableElement
from contoml.elements.tableheader import TableHeaderElement
from contoml.file.peekableit import PeekableIterator


class Entry:

    def __init__(self, names, table_element):
        self._table_element = table_element
        self._names = EntryName(names)

    @property
    def table_element(self):
        return self._table_element

    @property
    def name(self):
        """
        Returns an instnace of EntryName.
        """
        return self._names


class EntryName:
    """
    An entry's name.
    """

    def __init__(self, names):
        self._names = names

    @property
    def sub_names(self):
        return self._names

    def is_prefixed_with(self, names):
        if isinstance(names, EntryName):
            return self.is_prefixed_with(names.sub_names)

        for i, name in enumerate(names):
            if self._names[i] != name:
                return False
        return True

    def without_prefix(self, names):
        if isinstance(names, EntryName):
            return self.without_prefix(names.sub_names)

        for i, name in enumerate(names):
            if name != self._names[i]:
                return EntryName(self._names[i:])
        return EntryName(names=self.sub_names[len(names):])

    def __str__(self):
        return '.'.join(self.sub_names)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class AnonymousTableEntry(Entry):

    def __init__(self, table_element):
        Entry.__init__(self, ('',), table_element)


class TableEntry(Entry):

    def __init__(self, names, table_element):
        Entry.__init__(self, names=names, table_element=table_element)

class ArrayOfTablesEntry(Entry):

    def __init__(self, names, table_element):
        Entry.__init__(self, names=names, table_element=table_element)


def _anonymous_table(peekable_iter):
    """
    Returns the TableElement of the anonymous table, or raises a KeyError if not found.
    """
    try:
        first_table = peekable_iter.peek()
        if isinstance(first_table, TableElement):
            return first_table
        else:
            raise KeyError
    except StopIteration:
        raise KeyError


def _validate_file_elements(file_elements):
    # TODO
    pass


def extract(file_elements):
    """
    Outputs an ordered sequence of instances of Entry types.

    Elements start with an optional TableElement, followed by zero or more pairs of (TableHeaderElement, TableElement).
    """

    _validate_file_elements(file_elements)

    # An iterator over enumerate(the non-metadata) elements
    iterator = PeekableIterator((element_i, element) for (element_i, element) in enumerate(file_elements)
                                if element.type != elements.TYPE_METADATA)

    try:
        _, first_element = iterator.peek()
        if isinstance(first_element, TableElement):
            iterator.next()
            yield AnonymousTableEntry(first_element)
    except KeyError:
        pass
    except StopIteration:
        return

    for element_i, element in iterator:

        if not isinstance(element, TableHeaderElement):
            continue

        # If TableHeader of a regular table, return Table following it
        if not element.is_array_of_tables:
            table_element_i, table_element = next(iterator)
            yield TableEntry(names=element.names, table_element=table_element)

        # If TableHeader of an array of tables, do your thing
        else:
            table_element_i, table_element = next(iterator)
            yield ArrayOfTablesEntry(names=element.names, table_element=table_element)




"""
A regular expression based Lexer/tokenizer for TOML.
"""

from collections import namedtuple
import re
from contoml import tokens
from contoml.errors import TOMLError

TokenSpec = namedtuple('TokenSpec', ('type', 're'))

# Specs of all the valid tokens
_LEXICAL_SPECS = (
    TokenSpec(tokens.TYPE_COMMENT, re.compile(r'^(#.*)\n')),
    TokenSpec(tokens.TYPE_STRING, re.compile(r'^("(([^"]|\\")+?[^\\]|)")')),                       # Single line only
    TokenSpec(tokens.TYPE_MULTILINE_STRING, re.compile(r'^(""".*?""")', re.DOTALL)),
    TokenSpec(tokens.TYPE_LITERAL_STRING, re.compile(r"^('.*?')")),
    TokenSpec(tokens.TYPE_MULTILINE_LITERAL_STRING, re.compile(r"^('''.*?''')", re.DOTALL)),
    TokenSpec(tokens.TYPE_BARE_STRING, re.compile(r'^([A-Za-z0-9_-]+)')),
    TokenSpec(tokens.TYPE_DATE, re.compile(
        r'^([0-9]{4}-[0-9]{2}-[0-9]{2}(T[0-9]{2}:[0-9]{2}:[0-9]{2}(\.[0-9]*)?)?(([zZ])|((\+|-)[0-9]{2}:[0-9]{2}))?)')),
    TokenSpec(tokens.TYPE_WHITESPACE, re.compile(r'^( |\t)', re.DOTALL)),
    TokenSpec(tokens.TYPE_INTEGER, re.compile(r'^(((\+|-)[0-9_]+)|([1-9][0-9_]*))')),
    TokenSpec(tokens.TYPE_FLOAT,
              re.compile(r'^((((\+|-)[0-9_]+)|([1-9][0-9_]*))(\.[0-9_]+)?([eE](\+|-)?[0-9_]+)?)')),
    TokenSpec(tokens.TYPE_BOOLEAN, re.compile(r'^(true|false)')),
    TokenSpec(tokens.TYPE_OP_SQUARE_LEFT_BRACKET, re.compile(r'^(\[)')),
    TokenSpec(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, re.compile(r'^(\])')),
    TokenSpec(tokens.TYPE_OP_CURLY_LEFT_BRACKET, re.compile(r'^(\{)')),
    TokenSpec(tokens.TYPE_OP_CURLY_RIGHT_BRACKET, re.compile(r'^(\})')),
    TokenSpec(tokens.TYPE_OP_ASSIGNMENT, re.compile(r'^(=)')),
    TokenSpec(tokens.TYPE_OP_COMMA, re.compile(r'^(,)')),
    TokenSpec(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET, re.compile(r'^(\[\[)')),
    TokenSpec(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET, re.compile(r'^(\]\])')),
    TokenSpec(tokens.TYPE_OPT_DOT, re.compile(r'^(\.)')),
    TokenSpec(tokens.TYPE_NEWLINE, re.compile('^(\n|\r\n)')),
)

def _next_token_candidates(source):
    matches = []
    for token_spec in _LEXICAL_SPECS:
        match = token_spec.re.search(source)
        if match:
            matches.append(tokens.Token(token_spec.type, match.group(1)))
    return matches

def _choose_from_next_token_candidates(candidates):

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) > 1:
        # Return the maximal-munch with ties broken by natural order of token type.
        maximal_munch_length = max(len(token.source_substring) for token in candidates)
        maximal_munches = [token for token in candidates if len(token.source_substring) == maximal_munch_length]
        return sorted(maximal_munches)[0]   # Return the first in sorting by priority


def _munch_a_token(source):
    """
    Munches a single Token instance if it could recognize one at the beginning of the
    given source text, or None if no token type could be recognized.
    """
    candidates = _next_token_candidates(source)
    return _choose_from_next_token_candidates(candidates)


class LexerError(TOMLError):

    def __init__(self, message):
        self._message = message

    def __repr__(self):
        return self._message

    def __str__(self):
        return self._message


def tokenize(source):
    """
    Tokenizes the input TOML source into a stream of tokens.

    Raises a LexerError when it fails recognize another token while not at the end of the source.
    """

    # Newlines are going to be normalized to UNIX newlines.
    source = source.replace('\r\n', '\n')

    next_row = 1
    next_col = 1
    next_index = 0

    while next_index < len(source):

        new_token = _munch_a_token(source[next_index:])

        if not new_token:
            raise LexerError("failed to read the next token at ({}, {}): {}".format(
                next_row, next_col, source[next_index:]))

        # Set the col and row on the new token
        new_token = tokens.Token(new_token.type, new_token.source_substring, next_col, next_row)

        # Advance the index, row and col count
        next_index += len(new_token.source_substring)
        for c in new_token.source_substring:
            if c == '\n':
                next_row += 1
                next_col = 1
            else:
                next_col += 1

        yield new_token




from contoml.elements.table import TableElement
from contoml.errors import InvalidValueError
from contoml.file.freshtable import FreshTable


class ArrayOfTables(list):

    def __init__(self, parent, name, iterable=None):
        if iterable:
            list.__init__(self, iterable)
        self._name = name
        self._parent = parent

    def append(self, value):
        if isinstance(value, dict):
            table = FreshTable(parent=self, name=self._name, is_array=True)
            for k, v in value.items():
                table[k] = v
        else:
            raise InvalidValueError('Can only append a dict to an array of tables')

    def __getitem__(self, item):
        try:
            return list.__getitem__(self, item)
        except IndexError:
            if item == len(self):
                return FreshTable(parent=self, name=self._name, is_array=True)
            else:
                raise

    def append_fresh_table(self, fresh_table):
        list.append(self, fresh_table)
        if self._parent:
            self._parent.append_fresh_table(fresh_table)



from contoml import parser, lexer
from contoml.file import entries
from contoml.file.entries import AnonymousTableEntry, EntryName
from contoml.parser.tokenstream import TokenStream


def test_entry_extraction():
    text = open('sample.toml').read()
    elements = parser.parse_token_stream(TokenStream(lexer.tokenize(text)))

    e = tuple(entries.extract(elements))

    assert len(e) == 13
    assert isinstance(e[0], AnonymousTableEntry)


def test_entry_names():
    name_a = EntryName(('super', 'sub1'))
    name_b = EntryName(('super', 'sub1', 'sub2', 'sub3'))

    assert name_b.is_prefixed_with(name_a)
    assert name_b.without_prefix(name_a).sub_names == ('sub2', 'sub3')



from contoml.errors import InvalidValueError
from contoml.file.file import TOMLFile


def new():
    """
    Constructs a fresh empty TOML data structure.
    """
    from contoml.file.file import TOMLFile
    return TOMLFile([])


def loads(text):
    """
    Parses TOML text into a dict-like object and returns it.
    """
    from .parser import parse_token_stream
    from .lexer import tokenize as lexer
    from .parser.tokenstream import TokenStream
    from .file import TOMLFile

    tokens = lexer(text)
    elements = parse_token_stream(TokenStream(tokens))
    return TOMLFile(elements)


def load(file_path):
    """
    Parses a TOML file into a dict-like object and returns it.
    """
    return loads(open(file_path).read())


def dumps(value):
    """
    Dumps a data structure to TOML source code.

    The given value must be either a dict of dict values, a dict, or a TOML file constructed by this module.
    """

    if isinstance(value, TOMLFile):
        return value.dumps()

    if not isinstance(value, dict):
        raise InvalidValueError('Input must be a dict of dicts, or just a dict!')

    f = new()

    if all(isinstance(child, dict) for child in value.values()):
        for k, v in value.items():
            f[k] = v
    else:
        f[''] = value

    return f.dumps()


def dump(obj, file_path):
    """
    Dumps a data structure to the filesystem as TOML.

    The given value must be either a dict of dict values, a dict, or a TOML file constructed by this module.
    """
    with open(file_path, 'w') as fp:
        fp.write(dumps(obj))
